/*
 * MikroSDK - MikroE Software Development Kit
 * Copyright (c) 2019, MikroElektronika - www.mikroe.com
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*!
 * \file
 *
 */

#include "pressure5.h"

// ------------------------------------------------------------- PRIVATE MACROS 

#define PRESSURE5_DUMMY 0

#define I2C_COMMUNICATION                                                  0x00 
#define SPI_COMMUNICATION                                                  0x01


// ------------------------------------------------------------------ VARIABLES

static uint8_t coefficients[ 22 ] = {0};
static T_NVM_PAR   NVM_PAR;
static T_PAR_COEFF PAR;

// ---------------------------------------------- PRIVATE FUNCTION DECLARATIONS 

static void pressure5_i2c_write ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len );

static void pressure5_i2c_read ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len );

static void pressure5_spi_write ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len );

static void pressure5_spi_read ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len );

static void storage_coefficient ( void );

static uint8_t WAIT_DRDY ( pressure5_t *ctx, uint8_t TP );

// ------------------------------------------------ PUBLIC FUNCTION DEFINITIONS

void pressure5_cfg_setup ( pressure5_cfg_t *cfg )
{
    // Communication gpio pins 

    cfg->scl = HAL_PIN_NC;
    cfg->sda = HAL_PIN_NC;
    cfg->sck = HAL_PIN_NC;
    cfg->miso = HAL_PIN_NC;
    cfg->mosi = HAL_PIN_NC;
    cfg->cs = HAL_PIN_NC;
    
    // Additional gpio pins

    cfg->int_pin = HAL_PIN_NC;

    cfg->i2c_speed = I2C_STANDARD_SPEED; 
    cfg->i2c_address = PRESSURE5_DEVICE_SLAVE_ADDR_GND;
    cfg->spi_speed = SPI_MEDIUM; 
    cfg->spi_mode = SPI_MODE_0;
    
    //cfg->sel = PRESSURE5_MASTER_I2C;
    cfg->sel = PRESSURE5_MASTER_SPI;
}

PRESSURE5_RETVAL pressure5_init ( pressure5_t *ctx, pressure5_cfg_t *cfg )
{
    ctx->master_sel = cfg->sel;

    if ( ctx->master_sel == PRESSURE5_MASTER_I2C )
    {
        i2c_master_config_t i2c_cfg;

        i2c_master_configure_default( &i2c_cfg );
        i2c_cfg.speed    = cfg->i2c_speed;
        i2c_cfg.pins.scl = cfg->scl;
        i2c_cfg.pins.sda = cfg->sda;

        ctx->slave_address = cfg->i2c_address;

        if (  i2c_master_open( &ctx->i2c, &i2c_cfg ) != I2C_SUCCESS )
        {
            return PRESSURE5_INIT_ERROR;
        }

        ctx->read_f = pressure5_i2c_read;
        ctx->write_f = pressure5_i2c_write;
    }
    else
    {
        spi_master_config_t spi_cfg;

        spi_master_configure_default( &spi_cfg );
        spi_cfg.mode        = cfg->spi_mode;
        spi_cfg.speed       = cfg->spi_speed;
        spi_cfg.pins.sck    = cfg->sck;
        spi_cfg.pins.miso   = cfg->miso;
        spi_cfg.pins.mosi   = cfg->mosi;
        spi_cfg.chip_select = cfg->cs;

        if (  spi_master_open( &ctx->spi, &spi_cfg ) != SPI_SUCCESS )
        {   
            return  PRESSURE5_INIT_ERROR;
        }

         spi_master_set_dummy_data( &ctx->spi, PRESSURE5_DUMMY );
        
        ctx->read_f = pressure5_spi_read;
        ctx->write_f = pressure5_spi_write;

    }
    

    // Input pins
    
    digital_in_init( &ctx->int_pin, cfg->int_pin );
    

    return PRESSURE5_OK;
}

void pressure5_default_cfg ( pressure5_t *ctx )
{
    
    pressure5_write_byte( ctx, PRESSURE5_REG_OSR, PRESSURE5_OSR_PRESSURE_x1 | PRESSURE5_OSR_TEMP_x1 );
    pressure5_write_byte( ctx, PRESSURE5_REG_ODR, PRESSURE5_ODR_200 );
    
    pressure5_write_byte( ctx, PRESSURE5_REG_INT_CTRL, PRESSURE5_INTCFG_INT_OUTPUT_PUSH_PULL |
                                                 PRESSURE5_INTCFG_INT_LEVEL_ACTIVE_HIGH |
                                                 PRESSURE5_INTCFG_INT_LATCH_DISABLE |
                                                 PRESSURE5_INTCFG_FIFO_WTM_DISABLE |
                                                 PRESSURE5_INTCFG_FIFO_FULL_DISABLE |
                                                 PRESSURE5_INTCFG_DATA_READY_ENABLE );

    pressure5_write_byte( ctx, PRESSURE5_REG_IF_CONFIG, PRESSURE5_IFCFG_SPI_4_WIRE |
                                                  PRESSURE5_IFCFG_I2C_WATCHDOG_ENABLE |
                                                  PRESSURE5_IFCFG_I2C_WATCHDOG_LONG_40ms );
                                                  
    pressure5_write_byte( ctx, PRESSURE5_REG_POWER_CTRL,PRESSURE5_PCTRL_PRESSURE_SENSOR_ENABLE |
                                                  PRESSURE5_PCTRL_TEMP_SENSOR_ENABLE |
                                                  PRESSURE5_PCTRL_MODE_NORMAL );
        
    pressure5_update_coefficient( ctx );

}



void pressure5_generic_write ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    ctx->write_f( ctx, reg, data_buf, len ); 
}

void pressure5_generic_read ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    ctx->read_f( ctx, reg, data_buf, len );
}




void pressure5_write_byte( pressure5_t *ctx, uint8_t reg_addr, uint8_t data_in )
{

    pressure5_generic_write ( ctx, reg_addr, &data_in, 1);       

}

uint8_t pressure5_read_byte( pressure5_t *ctx, uint8_t reg_addr )
{
    uint8_t byte;
    
    pressure5_generic_read( ctx, reg_addr, &byte, 1 );

    return byte;
}

uint8_t pressure5_get_interrupt_state ( pressure5_t *ctx )
{
    return digital_in_read(&ctx->int_pin);
}

void pressure5_software_reset ( pressure5_t *ctx )
{
    pressure5_write_byte(ctx, PRESSURE5_REG_COMMAND, PRESSURE5_CMD_SOFTWARE_RESET);
}

uint32_t pressure5_get_raw_data ( pressure5_t *ctx, uint8_t data_addr )
{
    uint32_t data_raw;
    
    if(data_addr == PRESSURE5_TEMPERATURE_RAW_DATA)
    {
        WAIT_DRDY(ctx, 0x40);
    }
    else
    {
        WAIT_DRDY(ctx, 0x20);
    }
    
    data_raw = 0;

    data_raw = pressure5_read_byte(ctx, data_addr + 2);
    data_raw <<= 8;
    data_raw |=  pressure5_read_byte(ctx, data_addr + 1);
    data_raw <<= 8;
    data_raw |=  pressure5_read_byte(ctx, data_addr);
    
    return data_raw;
}


void pressure5_update_coefficient ( pressure5_t *ctx )
{ 
    pressure5_generic_read ( ctx, 0x31, coefficients, 21 );  
    storage_coefficient( );
}


float pressure5_get_temperature_data ( pressure5_t *ctx )
{
    uint32_t temp_raw;
    float partial_data1;
    float partial_data2;
    float calib_temp;
    
    temp_raw = pressure5_get_raw_data(ctx, PRESSURE5_TEMPERATURE_RAW_DATA);
    partial_data1 = (float)(temp_raw - PAR.T1);
    partial_data2 = (float)(partial_data1 * PAR.T2);
    
    calib_temp = partial_data2 + (partial_data1 * partial_data1) * PAR.T3;
    
    return calib_temp;
}

float pressure5_get_pressure_data ( pressure5_t *ctx )
{
    uint32_t pressure_raw;
    float temp_data;
    float comp_press;
    float data1;
    float data2;
    float data3;
    float data4;
    float out1;
    float out2;
    
    temp_data = pressure5_get_temperature_data( ctx );
    pressure_raw = pressure5_get_raw_data(ctx, PRESSURE5_PRESSURE_RAW_DATA);
    
    data1 = PAR.P6 * temp_data;
    data2 = PAR.P7 * (temp_data * temp_data);
    data3 = PAR.P8 * (temp_data * temp_data * temp_data);
    out1 = PAR.P5 + data1 + data2 + data3;
    
    data1 = PAR.P2 * temp_data;
    data2 = PAR.P3 * (temp_data * temp_data);
    data3 = PAR.P4 * (temp_data * temp_data * temp_data);
    out2 = (float)pressure_raw;
    out2 *= PAR.P1 + data1 + data2 + data3;
    
    data1 = (float)pressure_raw * (float)pressure_raw;
    data2 = PAR.P9 + PAR.P10 * temp_data;
    data3 = data1 * data2;
    data4 = data3 + ((float)pressure_raw * (float)pressure_raw * (float)pressure_raw) * PAR.P11;
    
    comp_press = out1 + out2 + data4;
    
    return comp_press / 100.0;
}

// ----------------------------------------------- PRIVATE FUNCTION DEFINITIONS

static void pressure5_i2c_write ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    i2c_master_transfer_t i2c_transfer;
    uint8_t tx_buf[ 256 ];
    uint8_t cnt;
    
    tx_buf[ 0 ] = reg;
    
    for ( cnt = 1; cnt <= len; cnt++ )
    {
        tx_buf[ cnt ] = data_buf[ cnt - 1 ]; 
    }
    
    i2c_transfer.slave_address = ctx->slave_address;
    i2c_transfer.buffer   = tx_buf;
    i2c_transfer.count    = len + 1;
    i2c_transfer.end_mode = I2C_STOP_MODE;

    i2c_master_start( &ctx->i2c );
    i2c_master_write( &ctx->i2c, &i2c_transfer );    
}

static void pressure5_i2c_read ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    i2c_master_transfer_t i2c_transfer;
    uint8_t tx_buf[ 1 ];

    tx_buf [ 0 ] = reg;
    
    i2c_transfer.slave_address = ctx->slave_address;
    i2c_transfer.buffer   = tx_buf;
    i2c_transfer.count    = 1;
    i2c_transfer.end_mode = I2C_RESTART_MODE;

    i2c_master_start( &ctx->i2c );
    i2c_master_write( &ctx->i2c, &i2c_transfer );
    
    i2c_transfer.buffer   = data_buf;
    i2c_transfer.count    = len;
    i2c_transfer.end_mode = I2C_STOP_MODE;

    i2c_master_read( &ctx->i2c, &i2c_transfer );
}

static void pressure5_spi_write ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    spi_master_transfer_data_t block;

    uint8_t tx_buf[ 265 ];
    uint8_t cnt;

    tx_buf[ 0 ] = reg;
    for ( cnt = 1; cnt <= len; cnt++ )
    {
        tx_buf[ cnt ] = data_buf[ cnt - 1 ]; 
    }
    
    block.tx_buffer = tx_buf;
    block.rx_buffer = 0;
    block.tx_length = len + 1;
    block.rx_length = 0;

    spi_master_start(&ctx->spi);
    spi_master_transfer(&ctx->spi, &block);
    spi_master_stop(&ctx->spi);    
}

static void pressure5_spi_read ( pressure5_t *ctx, uint8_t reg, uint8_t *data_buf, uint8_t len )
{
    spi_master_transfer_data_t block;

    uint8_t tx_buf[ 1 ];
    uint8_t rx_buf[ 265 ];
    uint8_t cnt;

    tx_buf[ 0 ] = reg | 0x80;
    
    block.tx_buffer = tx_buf;
    block.rx_buffer = rx_buf;
    block.tx_length = 1;
    block.rx_length = len + 2;
    
    spi_master_start(&ctx->spi);
    spi_master_transfer(&ctx->spi, &block);
    spi_master_stop(&ctx->spi);   

    for ( cnt = 0; cnt < len; cnt++ )
    {
        data_buf[ cnt ] = rx_buf [ cnt + 2 ];
    }
}

static void storage_coefficient( void )
{
    // Temperature 
    NVM_PAR.T1 = coefficients[ 1 ];
    NVM_PAR.T1 <<= 8;
    NVM_PAR.T1 |= coefficients[ 0 ];
    PAR.T1 = NVM_PAR.T1 / 0.00390625;
    
    NVM_PAR.T2 = coefficients[ 3 ];
    NVM_PAR.T2 <<= 8;
    NVM_PAR.T2 |= coefficients[ 2 ];
    PAR.T2 = NVM_PAR.T2 / 1073741824.0;
    
    NVM_PAR.T3 = coefficients[ 4 ];
    PAR.T3 = NVM_PAR.T3 / 281474976710656.0;
    
    // Pressure
    NVM_PAR.P1 = coefficients[ 6 ];
    NVM_PAR.P1 <<= 8;
    NVM_PAR.P1 |= coefficients[ 5 ];
    PAR.P1 = (NVM_PAR.P1 - 16384) / 1048576.0;
    
    NVM_PAR.P2 = coefficients[ 8 ];
    NVM_PAR.P2 <<= 8;
    NVM_PAR.P2 |= coefficients[ 7 ];
    PAR.P2 = (NVM_PAR.P2 - 16384) / 536870912.0;
    
    NVM_PAR.P3 = coefficients[ 9 ];
    PAR.P3 = NVM_PAR.P3 / 4294967296.0;
    
    NVM_PAR.P4 = coefficients[ 10 ];
    PAR.P4 = NVM_PAR.P4 / 137438953472.0;
    
    NVM_PAR.P5 = coefficients[ 12 ];
    NVM_PAR.P5 <<= 8;
    NVM_PAR.P5 |= coefficients[ 11 ];
    PAR.P5 = NVM_PAR.P5 / 0.125;
    
    NVM_PAR.P6 = coefficients[ 14 ];
    NVM_PAR.P6 <<= 8;
    NVM_PAR.P6 |= coefficients[ 13 ];
    PAR.P6 = NVM_PAR.P6 / 64.0;
     
    NVM_PAR.P7 = coefficients[ 15 ];
    PAR.P7 = NVM_PAR.P7 / 256.0;
    NVM_PAR.P8 = coefficients[ 16 ];
    PAR.P8 = NVM_PAR.P8 / 32768.0;
    
    NVM_PAR.P9 = coefficients[ 18 ];
    NVM_PAR.P9 <<= 8;
    NVM_PAR.P9 |= coefficients[ 17 ];
    PAR.P9 = NVM_PAR.P9 / 281474976710656.0;
    
    NVM_PAR.P10 = coefficients[ 19 ];
    PAR.P10 = NVM_PAR.P10 / 281474976710656.0;
    NVM_PAR.P11 = coefficients[ 20 ];
    PAR.P11 = NVM_PAR.P11 / 36893488147419103232.0;
}

static uint8_t WAIT_DRDY ( pressure5_t *ctx, uint8_t TP )
{
    uint8_t status;
    uint8_t cnt = 0;
    
    do
    {
        status = pressure5_read_byte(ctx, PRESSURE5_REG_INT_STATUS);
        cnt++;
        Delay_10ms();
        if ( cnt > 20 )
        {
            return 1;
        }
        
    }
    while ( (status & TP) == 0 );
    
    return 0;
}

// ------------------------------------------------------------------------- END

